import errno
import os
import typing
import warnings

import numpy as np
import pandas as pd

import audeer
import audinterface
import audobject

from opensmile.core.SMILEapi import (
    OpenSMILE,
    FrameMetaData,
)
from opensmile.core.config import config
from opensmile.core.define import (
    FeatureLevel,
    FeatureLevelResolver,
    FeatureSet,
    FeatureSetResolver,
)


class Smile(audinterface.Feature, audobject.Object):
    r"""OpenSMILE feature extractor.

    1. You can choose a pre-defined feature set by passing one of
       :class:`opensmile.FeatureSet`

    2. You can also provide a custom config file using the following
       template:

       .. code-block::

           [componentInstances:cComponentManager]
           instance[dataMemory].type=cDataMemory

           ;;; default source
           \{\cm[source{?}:source include config]}

           ;;; add components reading from reader.dmLevel=wave

           ;;; combine features
           [componentInstances:cComponentManager]
           instance[funcconcat].type=cVectorConcat
           [funcconcat:cVectorConcat]
           reader.dmLevel = <feat-1>;<feat-2>;...
           writer.dmLevel = features
           includeSingleElementFields = 1

           ;;; default sink
           \{\cm[sink{?}:include external sink]}

    .. note:: The following arguments are not serialized:

        * ``keep_nat``
        * ``loglevel``
        * ``logfile``
        * ``num_workers``
        * ``segment``
        * ``verbose``

        For more information see section on `hidden arguments`_.

    Args:
        feature_set: default feature set or path to a custom config file
        feature_level: default feature level or level name if a custom
            config file is used
        options: dictionary with optional script parameters
        loglevel: log level (0-5), the higher the number the more log
            messages are given
        logfile: if not ``None`` log messages will be stored to this file
        sampling_rate: sampling rate in Hz.
            If ``None`` it will call ``process_func`` with the actual
            sampling rate of the signal.
        channels: channel selection, see :func:`audresample.remix`
        mixdown: apply mono mix-down on selection
        resample: if ``True`` enforces given sampling rate by resampling
        segment: when a :class:`audinterface.Segment` object is provided,
            it will be used to find a segmentation of the input signal.
            Afterwards processing is applied to each segment
        keep_nat: if the end of segment is set to ``NaT`` do not replace
            with file duration in the result
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5 in case of
            multithreading and number of processors in case of
            multiprocessing
        multiprocessing: use multiprocessing instead of multithreading
        verbose: show debug messages

    Example:

    >>> sampling_rate = 16000
    >>> signal = np.zeros(sampling_rate)
    >>> smile = Smile(
    ...     feature_set=FeatureSet.ComParE_2016,
    ...     feature_level=FeatureLevel.Functionals,
    ... )
    >>> smile.process_signal(signal, sampling_rate).audspec_lengthL1norm_sma_range
    start   end
    0 days  0 days 00:00:01    0.0
    Name: audspec_lengthL1norm_sma_range, dtype: float32

    .. _`hidden arguments`: http://tools.pp.audeering.com/audobject/usage.html#hidden-arguments

    """  # noqa: E501
    @audobject.init_decorator(
        borrow={
            'sampling_rate': 'process',
            'channels': 'process',
            'mixdown': 'process',
            'resample': 'process',
        },
        hide=[
            'keep_nat',
            'logfile',
            'loglevel',
            'num_workers',
            'segment',
            'verbose',
        ],
        resolvers={
            'feature_set': FeatureSetResolver,
            'feature_level': FeatureLevelResolver,
        }
    )
    @audeer.deprecated_keyword_argument(
        deprecated_argument='num_channels',
        removal_version='0.13.0',
        new_argument='channels',
        mapping=lambda x: range(x),
    )
    def __init__(
            self,
            feature_set: typing.Union[
                str, FeatureSet
            ] = FeatureSet.ComParE_2016,
            feature_level: typing.Union[
                str, FeatureLevel
            ] = FeatureLevel.Functionals,
            *,
            options: dict = None,
            loglevel: int = 2,
            logfile: str = None,
            sampling_rate: int = None,
            channels: typing.Union[int, typing.Sequence[int]] = 0,
            mixdown: bool = False,
            resample: bool = False,
            segment: audinterface.Segment = None,
            keep_nat: bool = False,
            num_workers: typing.Optional[int] = 1,
            verbose: bool = False,
    ):

        self.feature_level = feature_level
        r"""Standard feature level or sink level in custom config file."""
        self.feature_set = feature_set
        r"""Standard feature set or path to custom config file"""
        self.options = options or {}
        r"""Dictionary with options"""
        self.logfile = audeer.safe_path(logfile) if logfile else None
        r"""Log file"""
        self.loglevel = loglevel
        r"""Log level"""
        self.verbose = verbose

        self._check_deltas_available()

        super().__init__(
            self._feature_names(),
            name='smile',
            params=None,
            process_func=self._extract,
            num_workers=num_workers,
            sampling_rate=sampling_rate,
            resample=resample,
            channels=channels,
            mixdown=mixdown,
            segment=segment,
            keep_nat=keep_nat,
            multiprocessing=True,
            verbose=verbose,
        )
        self.params = self.to_dict(flatten=True)

        self._check_deprecated()

    @property
    def default_config_root(self) -> str:
        r"""Return root directory with standard config files."""
        return os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            config.CONFIG_ROOT,
        )

    @property
    def config_name(self) -> str:
        r"""Return name of config file (without file extension)."""
        name, _ = os.path.splitext(os.path.basename(self.config_path))
        return name

    @property
    def config_path(self) -> str:
        r"""Return file path of config file."""

        if type(self.feature_set) is FeatureSet:
            config_path = os.path.join(
                self.default_config_root,
                self.feature_set.value + config.CONFIG_EXT,
            )
        else:
            config_path = audeer.safe_path(self.feature_set)

        if not os.path.exists(config_path):
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                config_path,
            )

        return config_path

    def _check_deltas_available(self):
        r"""Raise error if deltas are requested for GeMAPS family."""
        if self.feature_set in [
            FeatureSet.GeMAPS,
            FeatureSet.GeMAPSv01a,
            FeatureSet.GeMAPSv01b,
            FeatureSet.eGeMAPS,
            FeatureSet.eGeMAPSv01a,
            FeatureSet.eGeMAPSv01b,
            FeatureSet.eGeMAPSv02,
        ]:
            if self.feature_level == FeatureLevel.LowLevelDescriptors_Deltas:
                raise ValueError(
                    f"Feature level '{self.feature_level.name}' is not "
                    f"available for feature set '{self.feature_set.name}'."
                )

    def _check_deprecated(self):
        r"""Check if feature set is deprecated"""
        deprecated_feature_sets = {  # deprecated: recommended
            FeatureSet.GeMAPS: FeatureSet.GeMAPSv01b,
            FeatureSet.GeMAPSv01a: FeatureSet.GeMAPSv01b,
            FeatureSet.eGeMAPS: FeatureSet.eGeMAPSv01b,
            FeatureSet.eGeMAPSv01a: FeatureSet.eGeMAPSv02,
            FeatureSet.eGeMAPSv01b: FeatureSet.eGeMAPSv02,
        }
        if type(self.feature_set) is FeatureSet and \
                self.feature_set in deprecated_feature_sets:
            warnings.warn(
                f"Feature set '{self.feature_set}' is "
                f"deprecated, consider switching to "
                f"'{deprecated_feature_sets[self.feature_set]}'.",
                UserWarning,
            )

    def _extract(
            self,
            signal: np.ndarray,
            sampling_rate: int,
    ) -> (pd.TimedeltaIndex, pd.TimedeltaIndex, np.ndarray):
        r"""Run feature extraction."""

        signal = signal.copy()
        signal *= 32768
        signal = signal.astype(np.int16)

        ys = []
        starts = []
        ends = []

        for x in signal:

            y = []
            starts = []
            ends = []

            options = self._options()
            options['source'] = os.path.join(
                self.default_config_root,
                config.EXTERNAL_INPUT_CONFIG
            )
            options['sampleRate'] = sampling_rate
            options['nBits'] = 16

            smile = self._smile(options=options)
            smile.external_sink_set_callback_ex(
                config.EXTERNAL_OUTPUT_COMPONENT,
                Smile._sink_callback(y, starts, ends)
            )
            smile.external_audio_source_write_data(
                config.EXTERNAL_SOURCE_COMPONENT, bytes(x)
            )
            smile.external_audio_source_set_eoi(
                config.EXTERNAL_SOURCE_COMPONENT
            )
            smile.run()
            smile.free()

            if not y:
                warnings.warn(
                    UserWarning("Segment too short, filling with NaN.")
                )
                y.append(np.ones(self.num_features) * np.nan)
                starts.append(0)
                ends.append(signal.size / sampling_rate)

            starts = np.vstack(starts).squeeze()
            ends = np.vstack(ends).squeeze()
            if starts.shape:
                starts = pd.to_timedelta(starts, 's')
                ends = pd.to_timedelta(ends, 's')
            else:
                starts = pd.TimedeltaIndex([pd.to_timedelta(starts, 's')])
                ends = pd.TimedeltaIndex([pd.to_timedelta(ends, 's')])

            y = np.vstack(y)
            ys.append(y)

        return starts, ends, np.concatenate(ys, axis=1)

    def _feature_names(self) -> typing.List[str]:
        r"""Read feature names from config file."""
        options = self._options()
        options['source'] = os.path.join(
            self.default_config_root,
            config.EXTERNAL_INPUT_CONFIG
        )
        smile = self._smile(options=options)
        num_elements = smile.external_sink_get_num_elements(
            config.EXTERNAL_OUTPUT_COMPONENT
        )
        names = [
            smile.external_sink_get_element_name(
                config.EXTERNAL_OUTPUT_COMPONENT,
                idx) for idx in range(num_elements)
        ]
        smile.free()
        return names

    def _options(self) -> dict:
        r"""Fill options dictionary."""
        options = self.options.copy()
        options['sink'] = os.path.join(
            self.default_config_root,
            config.EXTERNAL_OUTPUT_SINGLE_CONFIG
        )
        if type(self.feature_level) is FeatureLevel:
            options['sinkLevel'] = self.feature_level.value
        else:
            options['sinkLevel'] = self.feature_level
        options['bufferModeRbConf'] = os.path.join(
            self.default_config_root,
            'shared/BufferModeRb.conf.inc'
        )
        if 'frameModeFunctionalsConf' not in options:
            options['frameModeFunctionalsConf'] = os.path.join(
                self.default_config_root,
                'shared/FrameModeFunctionals.conf.inc'
            )
        return options

    def _series_to_frame(
            self,
            series: pd.Series,
    ) -> pd.DataFrame:
        r"""Usually, we need to figure out start and end times from
        ``win_dur`` and ``hop_dur``. But since openSMILE provides segment
        times, we can skip this step and use them directly."""

        frames = [None] * len(series)
        if len(series.index.levels) == 3:
            for idx, ((file, start, end), values) in enumerate(series.items()):
                num = len(values[0])
                files = [file] * num
                starts = values[0] + start
                ends = values[1] + start
                # override first and last timestamp
                starts._values[0] = start
                ends._values[-1] = end
                values = values[2]
                index = pd.MultiIndex.from_arrays(
                    [
                        files,
                        starts,
                        ends,
                    ],
                    names=['file', 'start', 'end'],
                )
                frames[idx] = pd.DataFrame(
                    index=index,
                    data=values,
                    columns=self.column_names,
                )
        else:
            for idx, ((start, end), values) in enumerate(series.items()):
                starts = values[0] + start
                ends = values[1] + start
                # override first and last timestamp
                starts._values[0] = start
                ends._values[-1] = end
                values = values[2]
                index = pd.MultiIndex.from_arrays(
                    [
                        starts,
                        ends,
                    ],
                    names=['start', 'end'],
                )
                frames[idx] = pd.DataFrame(
                    index=index,
                    data=values,
                    columns=self.column_names,
                )
        return pd.concat(frames, axis='index')

    def _smile(self, options: dict) -> OpenSMILE:
        r"""Set up smile instance."""
        smile = OpenSMILE()
        smile.initialize(
            config_file=self.config_path,
            options=options,
            loglevel=self.loglevel,
            log_file=self.logfile,
            debug=self.verbose)
        return smile

    @staticmethod
    def _sink_callback(
        y: typing.List[np.ndarray],
        starts: typing.List[float],
        ends: typing.List[float]
    ) -> typing.Callable[[np.ndarray, FrameMetaData], None]:
        r"""Return callback where features are collected."""
        def callback(data: np.ndarray, meta: FrameMetaData):
            y.append(data.copy())
            starts.append(meta.time)
            ends.append(meta.time + meta.lengthSec)
        return callback

    def __call__(
            self,
            signal: np.ndarray,
            sampling_rate: int,
    ) -> np.ndarray:
        r"""Apply processing to signal.

        This function processes the signal **without** transforming the output
        into a :class:`pd.DataFrame`. Instead it will return the raw processed
        signal. However, if channel selection, mixdown and/or resampling
        is enabled, the signal will be first remixed and resampled if the
        input sampling rate does not fit the expected sampling rate.

        Args:
            signal: signal values
            sampling_rate: sampling rate in Hz

        Returns:
            Processed signal

        Raises:
            RuntimeError: if sampling rates do not match
            RuntimeError: if channel selection is invalid

        """
        # process functions returns (starts, values, values)
        # but we only want to return values here
        y = self.process(signal, sampling_rate)[2]
        # reshape to (channels, features, frames)
        y = y.T.reshape(self.num_channels, self.num_features, -1)
        return y



import configparser
from datetime import date
import os
from subprocess import check_output


config = configparser.ConfigParser()
config.read(os.path.join('..', 'setup.cfg'))

# Project -----------------------------------------------------------------

author = config['metadata']['author']
copyright = f'2020-{date.today().year} audEERING GmbH'
project = config['metadata']['name']
# The x.y.z version read from tags
try:
    version = check_output(['git', 'describe', '--tags', '--always'])
    version = version.decode().strip()
except Exception:
    version = '<unknown>'
title = '{} Documentation'.format(project)


# General -----------------------------------------------------------------

master_doc = 'index'
extensions = []
source_suffix = '.rst'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']
pygments_style = None
extensions = [
    'jupyter_sphinx',  # executing code blocks
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',  # support for Google-style docstrings
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosectionlabel',
    'sphinx_autodoc_typehints',
    'sphinx_copybutton',  # for "copy to clipboard" buttons
]

napoleon_use_ivar = True  # List of class attributes
autodoc_inherit_docstrings = False  # disable docstring inheritance

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'pandas': ('https://pandas-docs.github.io/pandas-docs-travis/', None),
}

# Do not copy prompot output
copybutton_prompt_text = r'>>> |\.\.\. |\$ '
copybutton_prompt_is_regexp = True

# HTML --------------------------------------------------------------------

html_theme = 'sphinx_audeering_theme'
html_theme_options = {
    'display_version': True,
    'footer_links': False,
    'logo_only': False,
    'wide_pages': ['usage'],
}
html_context = {
    'display_github': True,
}
html_title = title

# Linkcheck ---------------------------------------------------------------

linkcheck_ignore = [
    'https://sail.usc.edu/',
    'http://sphinx-doc.org/',
]



import audeer
import os
import glob
import sys

import pytest
import pandas as pd
import audiofile as af


pytest.ROOT = os.path.dirname(os.path.realpath(__file__))
pytest.WAV_FILE = os.path.join(pytest.ROOT, 'test.wav')
pytest.WAV_ARRAY, pytest.WAV_SR = af.read(pytest.WAV_FILE, always_2d=True)
pytest.FRAME_LIST_STARTS = pd.to_timedelta(['1.0s', '3.0s', '4.0s'])
pytest.FRAME_LIST_ENDS = pd.to_timedelta(['1.5s', '3.5s', '5.0s'])
pytest.CONFIG_FILE = os.path.join(pytest.ROOT, 'test.conf')

if sys.platform == "win32":  # pragma: no cover
    platform = 'win'
elif sys.platform == "darwin":  # pragma: no cover
    platform = 'osx'
else:  # pragma: no cover
    platform = 'linux'

pytest.SMILEXTRACT = audeer.safe_path(
    os.path.join(
        pytest.ROOT, '..', 'opensmile', 'core', 'bin', platform, 'SMILExtract'
    )
)


@pytest.fixture(scope='session', autouse=True)
def fixture_clean_session():
    def clean():
        path = os.path.join(pytest.ROOT, '..', '.coverage.*')
        for file in glob.glob(path):
            os.remove(file)
    clean()
    yield
    clean()



from ctypes import cdll, Structure, c_char, c_char_p, c_int, c_long, \
    c_void_p, c_float, c_double, CFUNCTYPE, POINTER, cast, byref
from typing import Any, Callable, Dict, List, Optional
import json
import os
import numpy as np
import sys

root = os.path.dirname(os.path.realpath(__file__))

if sys.platform == "win32":  # pragma: no cover
    smileapi_path = os.path.join(
        root, 'bin', 'win', 'SMILEapi.dll'
    )
elif sys.platform == "darwin":  # pragma: no cover
    smileapi_path = os.path.join(
        root, 'bin', 'osx', 'libSMILEapi.dylib'
    )
else:  # pragma: no cover
    smileapi_path = os.path.join(
        root, 'bin', 'linux', 'libSMILEapi.so',
    )
smileapi = cdll.LoadLibrary(smileapi_path)

# definitions from smileComponent.hpp
CMSG_textLen = 64
CMSG_typenameLen = 32
CMSG_nUserData = 8


class ComponentMessage(Structure):  # pragma: no cover
    """
    An openSMILE component message.
    """

    _fields_ = [
        ("_msgtype", c_char * CMSG_typenameLen),
        ("_msgname", c_char * CMSG_typenameLen),
        ("_sender", c_char_p),
        ("smileTime", c_double),
        ("userTime1", c_double),
        ("userTime2", c_double),
        ("readerTime", c_double),
        ("msgid", c_int),
        ("floatData", c_double * CMSG_nUserData),
        ("intData", c_int * CMSG_nUserData),
        ("_msgtext", c_char * CMSG_textLen),
        ("userflag1", c_int),
        ("userflag2", c_int),
        ("userflag3", c_int),
        ("custData", c_void_p),
        ("custData2", c_void_p),
        ("custDataSize", c_int),
        ("custData2Size", c_int),
        ("custDataType", c_int),
        ("custData2Type", c_int)
    ]

    @property
    def msgtype(self):
        return self._msgtype.decode("utf-8")

    @property
    def msgname(self):
        return self._msgname.decode("utf-8")

    @property
    def sender(self):
        return self._sender.decode("utf-8")

    @property
    def msgtext(self):
        return self._msgtext.decode("utf-8")

    def unpack_json(self):
        """
        Unpacks a component message that wraps a JSON object and returns the
        JSON data as a dictionary.
        """
        if self.msgtype != "_CONTAINER" or self.msgname != "jsonObject":
            raise ValueError("Message does not contain JSON data")
        cust_data = cast(self.custData, c_char_p)
        if cust_data:
            return json.loads(cust_data.value.decode("ascii"))
        else:
            return None

    def __str__(self):
        return "type: {}, name: {}, sender: {}, msgtext: {}".format(
            self.msgtype,
            self.msgname,
            self.sender,
            self.msgtext)


# Success and error return codes
SMILE_SUCCESS = 0  # success
SMILE_FAIL = 1  # generic error
SMILE_INVALID_ARG = 2  # an invalid argument was passed
SMILE_INVALID_STATE = 3  # openSMILE was in an invalid state for
# the called function
SMILE_COMP_NOT_FOUND = 4  # component instance was not found
SMILE_LICENSE_FAIL = 5  # license validation check failed
SMILE_CONFIG_PARSE_FAIL = 6  # configuration could not be loaded
SMILE_CONFIG_INIT_FAIL = 7  # configuration could not be initialized
SMILE_NOT_WRITTEN = 8  # data could not be written to a
# cExternalSource/cExternalAudioSource component

# openSMILE states
SMILE_UNINITIALIZED = 0  # no configuration has been loaded yet
SMILE_INITIALIZED = 1  # a configuration has been loaded
SMILE_RUNNING = 2  # openSMILE is running
# SMILE_PAUSED,
SMILE_ENDED = 3  # openSMILE has finished


class FrameMetaData(Structure):  # pragma: no cover
    _fields_ = [
        ("vIdx", c_long),
        ("time", c_double),
        ("period", c_double),
        ("lengthSec", c_double)
    ]

    def __str__(self):
        return "vIdx: {}, time: {}, period: {}, lengthSec: {}".format(
            self.vIdx,
            self.time,
            self.period,
            self.lengthSec)


StateChangedCallback = CFUNCTYPE(c_void_p, c_int, c_void_p)
# TODO: return type bool does not exist in C!
ExternalSinkCallback = CFUNCTYPE(c_int, POINTER(c_float), c_long, c_void_p)
ExternalSinkCallbackEx = CFUNCTYPE(c_int, POINTER(c_float), c_long, c_long,
                                   POINTER(FrameMetaData), c_void_p)
ExternalMessageInterfaceCallback = CFUNCTYPE(c_int, POINTER(ComponentMessage),
                                             c_void_p)
ExternalMessageInterfaceJsonCallback = CFUNCTYPE(c_int, c_char_p, c_void_p)

smileapi.smile_new.argtypes = []
smileapi.smile_new.restype = c_void_p
smileapi.smile_initialize.argtypes = [c_void_p, c_char_p, c_int, c_void_p,
                                      c_int, c_int, c_int, c_void_p]
smileapi.smile_initialize.restype = c_int
smileapi.smile_run.argtypes = [c_void_p]
smileapi.smile_run.restype = c_int
smileapi.smile_abort.argtypes = [c_void_p]
smileapi.smile_abort.restype = c_int
smileapi.smile_reset.argtypes = [c_void_p]
smileapi.smile_reset.restype = c_int
smileapi.smile_get_state.argtypes = [c_void_p]
smileapi.smile_get_state.restype = c_int
smileapi.smile_set_state_callback.argtypes = [c_void_p, StateChangedCallback,
                                              c_void_p]
smileapi.smile_set_state_callback.restype = c_int
smileapi.smile_free.argtypes = [c_void_p]
smileapi.smile_free.restype = None
smileapi.smile_extsource_write_data.argtypes = [c_void_p, c_char_p,
                                                POINTER(c_float), c_int]
smileapi.smile_extsource_write_data.restype = c_int
smileapi.smile_extsource_set_external_eoi.argtypes = [c_void_p, c_char_p]
smileapi.smile_extsource_set_external_eoi.restype = c_int
smileapi.smile_extaudiosource_write_data.argtypes = [c_void_p, c_char_p,
                                                     c_void_p, c_int]
smileapi.smile_extaudiosource_write_data.restype = c_int
smileapi.smile_extaudiosource_set_external_eoi.argtypes = [c_void_p, c_char_p]
smileapi.smile_extaudiosource_set_external_eoi.restype = c_int
smileapi.smile_extsink_set_data_callback.argtypes = [c_void_p, c_char_p,
                                                     ExternalSinkCallback,
                                                     c_void_p]
smileapi.smile_extsink_set_data_callback.restype = c_int
smileapi.smile_extsink_set_data_callback_ex.argtypes = [c_void_p, c_char_p,
                                                        ExternalSinkCallbackEx,
                                                        c_void_p]
smileapi.smile_extsink_set_data_callback_ex.restype = c_int
smileapi.smile_extsink_get_num_elements.argtypes = [c_void_p, c_char_p,
                                                    POINTER(c_long)]
smileapi.smile_extsink_get_num_elements.restype = c_int
smileapi.smile_extsink_get_element_name.argtypes = [c_void_p, c_char_p, c_long,
                                                    POINTER(c_char_p)]
smileapi.smile_extsink_get_element_name.restype = c_int
smileapi.smile_extmsginterface_set_msg_callback.argtypes = \
    [c_void_p, c_char_p, ExternalMessageInterfaceCallback, c_void_p]
smileapi.smile_extmsginterface_set_msg_callback.restype = c_int
smileapi.smile_extmsginterface_set_json_msg_callback.argtypes = \
    [c_void_p, c_char_p, ExternalMessageInterfaceJsonCallback, c_void_p]
smileapi.smile_extmsginterface_set_json_msg_callback.restype = c_int
smileapi.smile_error_msg.argtypes = [c_void_p]
smileapi.smile_error_msg.restype = c_char_p


def c_char_p_arr(x):
    arr = (c_char_p * (len(x) + 1))()
    arr[:-1] = x
    arr[-1] = None
    return arr


class OpenSmileException(Exception):  # pragma: no cover
    """
    Exception thrown for internal openSMILE errors.
    """

    def __init__(self, code: int, message: Optional[str] = None):
        self.code = code
        self.message = message

    def __str__(self):
        if self.message:
            return "Code: {}, Message: {}".format(self.code, self.message)
        else:
            return "Code: {}".format(self.code)


class OpenSMILE(object):  # pragma: no cover
    """
    The main class implementing the interface to openSMILE.
    """

    def __init__(self):
        self._smileobj = None

    def initialize(self, config_file: str, options: Dict[str, Any] = None,
                   loglevel: int = 2, debug: bool = False,
                   console_output: bool = False, log_file: str = None):
        """
        Initializes openSMILE with the provided config file and command-line
         options.
        """
        self._smileobj = smileapi.smile_new()
        if self._smileobj is None:
            raise OpenSmileException(SMILE_FAIL,
                                     "could not create new SMILEapi object")
        options_flat = list(
            map(lambda v: bytes(str(v), "ascii"), sum(options.items(), ())))
        options_char_arr = c_char_p_arr(options_flat)
        log_file = bytes(log_file, "ascii") if log_file else int(0)
        self._check_smile_result(smileapi.smile_initialize(self._smileobj,
                                                           bytes(config_file,
                                                                 "ascii"),
                                                           len(options),
                                                           options_char_arr,
                                                           loglevel,
                                                           int(debug),
                                                           int(console_output),
                                                           log_file))
        self._callbacks = []

    def external_source_write_data(self, component_name: str,
                                   data: np.ndarray) -> bool:
        """
        Writes a data buffer to the specified instance of a cExternalSource
        component.

        Returns True if the data was written successfully, otherwise returns
        False (e.g. if the internal buffer of the component is full).
        """
        if len(data.shape) != 1:
            raise ValueError("data parameter must have exactly one dimension")
        if data.dtype.name != "float32":
            raise ValueError("data parameter must have dtype float32")
        data_p = data.ctypes.data_as(POINTER(c_float))
        result = smileapi.smile_extsource_write_data(self._smileobj,
                                                     bytes(component_name,
                                                           "ascii"), data_p,
                                                     len(data))
        if result == SMILE_SUCCESS:
            return True
        elif result == SMILE_NOT_WRITTEN:
            return False
        else:
            self._check_smile_result(result)

    def external_source_set_eoi(self, component_name: str):
        """
        Signals the end of the input for the specified cExternalSource
        component instance.
        Attempts to write more data to the component after calling this
        method will fail.

        Returns True if the end-of-input signal was set successfully,
        otherwise False.
        """
        self._check_smile_result(
            smileapi.smile_extsource_set_external_eoi(self._smileobj,
                                                      bytes(component_name,
                                                            "ascii")))

    def external_audio_source_write_data(self, component_name: str,
                                         data: bytes) -> bool:
        """
        Writes a data buffer to the specified instance of a
        cExternalAudioSource component.
        The data must match the specified data format for the component
        (sample size, number of channels, etc.).

        Returns True if the data was written successfully, otherwise
        returns False
        (e.g. if the internal buffer of the component is full).
        """
        result = smileapi.smile_extaudiosource_write_data(self._smileobj,
                                                          bytes(component_name,
                                                                "ascii"), data,
                                                          len(data))
        if result == SMILE_SUCCESS:
            return True
        elif result == SMILE_NOT_WRITTEN:
            return False
        else:
            self._check_smile_result(result)

    def external_audio_source_set_eoi(self, component_name: str):
        """
        Signals the end of the input for the specified cExternalAudioSource
        component instance.
        Attempts to write more data to the component after calling this method
        will fail.

        Returns True if the end-of-input signal was set successfully,
        otherwise False.
        """
        self._check_smile_result(
            smileapi.smile_extaudiosource_set_external_eoi(self._smileobj,
                                                           bytes(
                                                               component_name,
                                                               "ascii")))

    def external_sink_set_callback(self, component_name: str,
                                   callback: Callable[[np.ndarray], None]):
        """
        Sets the callback function for the specified cExternalSink
        component instance.
        The function will get called whenever another openSMILE component
        writes data to the cExternalSink component.
        """

        def internal_callback(data, vector_size, param):
            numpy_array = np.ctypeslib.as_array(data, shape=(vector_size,))
            callback(numpy_array)
            return 1

        cb = ExternalSinkCallback(internal_callback)
        # we need to keep a reference to any callback objects as otherwise
        # they may get garbage-collected
        self._callbacks.append(cb)
        self._check_smile_result(
            smileapi.smile_extsink_set_data_callback(self._smileobj,
                                                     bytes(component_name,
                                                           "ascii"), cb, None))

    def external_sink_set_callback_ex(self, component_name: str,
                                      callback: Callable[[np.ndarray], None]):
        """
        Sets the extended callback function for the specified
        cExternalSink component instance.
        The function will get called whenever another openSMILE component
        writes data to the cExternalSink component.
        """

        def internal_callback_ex(data, nt, n, meta: POINTER(FrameMetaData), _):
            numpy_array = np.ctypeslib.as_array(data, shape=(nt, n))
            callback(numpy_array, meta.contents)
            return 1

        cb = ExternalSinkCallbackEx(internal_callback_ex)
        # we need to keep a reference to any callback objects as otherwise
        # they may get garbage-collected
        self._callbacks.append(cb)
        self._check_smile_result(
            smileapi.smile_extsink_set_data_callback_ex(self._smileobj,
                                                        bytes(component_name,
                                                              "ascii"), cb,
                                                        None))

    def external_sink_get_num_elements(self, component_name: str) -> int:
        num_elements = c_long()
        self._check_smile_result(
            smileapi.smile_extsink_get_num_elements(self._smileobj,
                                                    bytes(component_name,
                                                          "ascii"),
                                                    byref(num_elements)))
        return num_elements.value

    def external_sink_get_element_name(self, component_name: str,
                                       idx: int) -> str:
        element_name = c_char_p()
        self._check_smile_result(
            smileapi.smile_extsink_get_element_name(self._smileobj,
                                                    bytes(component_name,
                                                          "ascii"), idx,
                                                    byref(element_name)))
        return element_name.value.decode("ascii")

    def external_message_interface_set_callback(self, component_name: str,
                                                callback: Callable[
                                                    [ComponentMessage], None]):
        """
        Sets the callback function for the specified cExternalMessageInterface
        component instance.
        The function will get called whenever the component receives a message.
        """

        # we need to keep a reference to any callback objects as otherwise
        # they may get garbage-collected
        def internal_callback(message: POINTER(ComponentMessage), param):
            callback(message.contents)
            return 1

        cb = ExternalMessageInterfaceCallback(internal_callback)
        self._callbacks.append(cb)
        self._check_smile_result(
            smileapi.smile_extmsginterface_set_msg_callback(self._smileobj,
                                                            bytes(
                                                                component_name,
                                                                "ascii"), cb,
                                                            None))

    def external_message_interface_set_json_callback(self, component_name: str,
                                                     callback: Callable[
                                                         [Dict], None]):
        """
        Sets the callback function for the specified cExternalMessageInterface
        component instance.
        The function will get called whenever the component receives a message.
        """

        # we need to keep a reference to any callback objects as otherwise
        # they may get garbage-collected
        def internal_callback(json_message: bytes, param):
            callback(json.loads(json_message.decode("ascii")))
            return 1

        cb = ExternalMessageInterfaceJsonCallback(internal_callback)
        self._callbacks.append(cb)
        self._check_smile_result(
            smileapi.smile_extmsginterface_set_json_msg_callback(
                self._smileobj, bytes(component_name, "ascii"), cb, None))

    def run(self):
        """
        Starts processing and blocks until finished.
        """
        self._check_smile_result(smileapi.smile_run(self._smileobj))

    def abort(self):
        """
        Requests abortion of the current run.

        Note that openSMILE does not immediately stop after this function
        returns. It might continue to run for a short while until the run
        method returns.
        """
        self._check_smile_result(smileapi.smile_abort(self._smileobj))

    def reset(self):
        """
        Resets the internal state of openSMILE after a run has finished or
        was aborted. After resetting, you may call 'run' again without the
        need to call 'initialize' first. You must re-register any
        cExternalSink/cExternalMessageInterface callbacks, though.
        """
        self._check_smile_result(smileapi.smile_reset(self._smileobj))

    def free(self):
        """
        Frees any internal resources allocated by openSMILE.
        """
        if self._smileobj is not None:
            smileapi.smile_free(self._smileobj)
            self._smileobj = None

    def _check_smile_result(self, result: int):
        if result != SMILE_SUCCESS:
            message = smileapi.smile_error_msg(self._smileobj)
            if message is None or len(message) == 0:
                raise OpenSmileException(result)
            else:
                raise OpenSmileException(result, message.decode("ascii"))

    @staticmethod
    def process(config_file: str, options: Dict[str, Any],
                inputs: Dict[str, np.ndarray], outputs: List[str]) \
            -> Dict[str, np.ndarray]:
        """
        Runs the specified config file on a set of input data buffers and
        returns the specified set of output buffers.
        """
        opensmile = OpenSMILE()
        opensmile.initialize(config_file, options)

        for (input, data) in inputs.items():
            if not opensmile.external_source_write_data(input, data):
                raise Exception(
                    "Could not write input data to component '{}'".format(
                        input))
            opensmile.external_source_set_eoi(input)

        output_data = {}

        for output in outputs:
            def callback(data: np.ndarray):
                if output not in output_data:
                    output_data[output] = np.copy(data)
                else:
                    output_data[output] = np.vstack(
                        (output_data[output], data))

            opensmile.external_sink_set_callback(output, callback)

        opensmile.run()
        opensmile.free()
        return output_data



from opensmile.core.config import config
from opensmile.core.define import (
    FeatureSet,
    FeatureLevel,
)
from opensmile.core.smile import (
    Smile,
)


__all__ = []


__version__ = 'unknown'

# Dynamically get the version of the installed module
try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:  # pragma: no cover
    pkg_resources = None  # pragma: no cover
finally:
    del pkg_resources



import enum
import typing

import audobject


class FeatureSet(enum.Enum):
    r""" Enumeration of standard feature sets.

    The following feature sets are available:

    * :attr:`ComParE_2016`
    * :attr:`GeMAPS`, deprecated alias for :attr:`GeMAPSv01a`
    * :attr:`GeMAPSv01a`
    * :attr:`GeMAPSv01b`
    * :attr:`eGeMAPS`, deprecated alias for :attr:`eGeMAPSv01a`
    * :attr:`eGeMAPSv01a`
    * :attr:`eGeMAPSv01b`
    * :attr:`eGeMAPSv02`
    * :attr:`emobase`

    For references, see the papers on:

    * `ComParE 2016`_
    * GeMAPS_
    * eGeMAPS_
    * emobase

    .. _ComParE 2016:
        http://www.tangsoo.de/documents/Publications/Schuller16-TI2.pdf
    .. _GeMAPS:
        https://sail.usc.edu/publications/files/eyben-preprinttaffc-2015.pdf
    .. _eGeMAPS:
        https://sail.usc.edu/publications/files/eyben-preprinttaffc-2015.pdf

    """
    ComParE_2016 = 'compare/ComParE_2016'
    GeMAPS = 'gemaps/v01a/GeMAPSv01a'  # legacy
    GeMAPSv01a = 'gemaps/v01a/GeMAPSv01a'
    GeMAPSv01b = 'gemaps/v01b/GeMAPSv01b'
    eGeMAPS = 'egemaps/v01a/eGeMAPSv01a'   # legacy
    eGeMAPSv01a = 'egemaps/v01a/eGeMAPSv01a'
    eGeMAPSv01b = 'egemaps/v01b/eGeMAPSv01b'
    eGeMAPSv02 = 'egemaps/v02/eGeMAPSv02'
    emobase = 'emobase/emobase'


class FeatureSetResolver(audobject.resolver.Base):
    r"""Custom value resolver for :class:`opensmile.FeatureSet`."""

    def decode(self, value: str) -> typing.Union[str, FeatureSet]:
        if value in FeatureSet.__members__:
            value = FeatureSet[value]
        return value

    def encode(self, value: typing.Union[str, FeatureSet]) -> str:
        if isinstance(value, FeatureSet):
            value = str(value).split('.')[-1]
        return value

    def encode_type(self):
        return str


class FeatureLevel(enum.Enum):
    r""" Enumeration of standard feature levels.

    * :attr:`LowLevelDescriptors` - low-level descriptors (LLD) calculated
      over a sliding window
    * :attr:`LowLevelDescriptors_Deltas` - Delta regression of LLDs
    * :attr:`Functionals` - statistical functionals mapping variable series of
      LLDs to static values

    For more information see https://mediatum.ub.tum.de/doc/1082431/1082431.pdf

    """
    LowLevelDescriptors = 'lld'
    LowLevelDescriptors_Deltas = 'lld_de'
    Functionals = 'func'


class FeatureLevelResolver(audobject.resolver.Base):
    r"""Custom value resolver for :class:`opensmile.FeatureLevel`."""

    def decode(self, value: str) -> typing.Union[str, FeatureLevel]:
        if value in FeatureLevel.__members__:
            value = FeatureLevel[value]
        return value

    def encode(self, value: typing.Union[str, FeatureLevel]) -> str:
        if isinstance(value, FeatureLevel):
            value = str(value).split('.')[-1]
        return value

    def encode_type(self):
        return str



import os

import numpy as np
import pandas as pd
import pytest

import audeer
import audiofile
import audobject

import opensmile


deprecated_feature_sets = [  # deprecated
    opensmile.FeatureSet.GeMAPS,
    opensmile.FeatureSet.GeMAPSv01a,
    opensmile.FeatureSet.eGeMAPS,
    opensmile.FeatureSet.eGeMAPSv01a,
    opensmile.FeatureSet.eGeMAPSv01b,
]

gemaps_family = [  # no deltas
    opensmile.FeatureSet.GeMAPS,
    opensmile.FeatureSet.GeMAPSv01a,
    opensmile.FeatureSet.GeMAPSv01b,
    opensmile.FeatureSet.eGeMAPS,
    opensmile.FeatureSet.eGeMAPSv01a,
    opensmile.FeatureSet.eGeMAPSv01b,
    opensmile.FeatureSet.eGeMAPSv02,
]


@pytest.mark.parametrize('x,sr,num_channels,feature_set,feature_level', [
    (
        pytest.WAV_ARRAY,
        pytest.WAV_SR,
        3,
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.LowLevelDescriptors,
    ),
    (
        pytest.WAV_ARRAY,
        pytest.WAV_SR,
        5,
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.Functionals,
    ),
])
def test_channels(x, sr, num_channels, feature_set, feature_level):

    x = np.repeat(x, num_channels, axis=0)

    # create feature extractor with mixdown

    fex = opensmile.Smile(feature_set, feature_level, mixdown=True)
    fex = audobject.from_yaml_s(fex.to_yaml_s())
    assert isinstance(fex, opensmile.Smile)

    y_mono = fex.process_signal(x, sr)

    # create feature extractor for multiple channels

    fex = opensmile.Smile(
        feature_set, feature_level, channels=range(num_channels),
    )
    fex = audobject.from_yaml_s(fex.to_yaml_s())
    assert isinstance(fex, opensmile.Smile)

    y = fex.process_signal(x, sr)

    # assertions

    assert y_mono.shape[0] == y.shape[0]
    assert y_mono.shape[1] * len(fex.process.channels) == y.shape[1]
    for c in range(num_channels):
        np.testing.assert_equal(
            y.values[:, c * fex.num_features:(c + 1) * fex.num_features],
            y_mono.values,
        )


@pytest.mark.parametrize('config,level', [
    (
        pytest.CONFIG_FILE,
        'lld',
    ),
    (
        pytest.CONFIG_FILE,
        'func',
    ),
    pytest.param(
        'invalid.conf',
        'func',
        marks=pytest.mark.xfail(raises=FileNotFoundError),
    ),
])
def test_custom(config, level):

    # create feature extractor

    fex = opensmile.Smile(config, level)
    fex = audobject.from_yaml_s(fex.to_yaml_s())
    assert isinstance(fex, opensmile.Smile)

    # extract from file

    y_file = fex.process_file(pytest.WAV_FILE)

    # extract from array

    x, sr = audiofile.read(pytest.WAV_FILE)
    y_array = fex.process_signal(x, sr, file=pytest.WAV_FILE)

    # assertions

    assert fex.config_name == audeer.basename_wo_ext(config)
    assert fex.config_path == audeer.safe_path(config)
    assert fex.num_features == len(fex.feature_names)
    assert fex.feature_names == y_file.columns.to_list()
    pd.testing.assert_frame_equal(y_file, y_array)


@pytest.mark.parametrize(
    'feature_set',
    [
        x for x in opensmile.FeatureSet
    ]
)
@pytest.mark.parametrize(
    'feature_level',
    [
        x for x in opensmile.FeatureLevel
    ]
)
def test_default(tmpdir, feature_set, feature_level):

    deltas = feature_level == opensmile.FeatureLevel.LowLevelDescriptors_Deltas

    if (feature_set in gemaps_family) and deltas:

        # deltas not available

        with pytest.raises(ValueError):
            opensmile.Smile(feature_set, feature_level)

    else:

        # create feature extractor

        if feature_set in deprecated_feature_sets:
            with pytest.warns(UserWarning):
                fex = opensmile.Smile(feature_set, feature_level)
                fex = audobject.from_yaml_s(fex.to_yaml_s())
                assert isinstance(fex, opensmile.Smile)
        else:
            fex = opensmile.Smile(feature_set, feature_level)
            fex = audobject.from_yaml_s(fex.to_yaml_s())
            assert isinstance(fex, opensmile.Smile)

        # extract features from file

        y = fex.process_file(pytest.WAV_FILE)

        # run SMILExtract from same file

        source_config_file = os.path.join(
            fex.default_config_root,
            opensmile.config.FILE_INPUT_CONFIG,
        )
        if feature_set in gemaps_family:
            sink_config_file = os.path.join(
                fex.default_config_root,
                opensmile.config.FILE_OUTPUT_CONFIG_NO_LLD_DE,
            )
        else:
            sink_config_file = os.path.join(
                fex.default_config_root,
                opensmile.config.FILE_OUTPUT_CONFIG,
            )
        output_file = os.path.join(tmpdir, f'{feature_level.value}.csv')
        command = f'{pytest.SMILEXTRACT} ' \
                  f'-C {fex.config_path} ' \
                  f'-source {source_config_file} ' \
                  f'-I {pytest.WAV_FILE} ' \
                  f'-sink {sink_config_file} ' \
                  f'-{feature_level.value}_csv_output {output_file}'
        os.system(command)

        # read output of SMILExtract and compare

        df = pd.read_csv(output_file, sep=';')
        np.testing.assert_allclose(df.values[:, 1:],
                                   y.values,
                                   rtol=1e-6,
                                   atol=0)
        assert fex.num_features == len(df.columns) - 1
        assert fex.feature_names == list(df.columns[1:])


@pytest.mark.parametrize('num_files', [
    1, 5
])
@pytest.mark.parametrize('feature_set,feature_level', [
    (
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.LowLevelDescriptors
    ),
    (
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.Functionals
    ),
])
@pytest.mark.parametrize(
    'num_workers',
    [
        1, 5, None,
    ],
)
def test_files(num_files, feature_set, feature_level, num_workers):

    # create feature extractor

    fex = opensmile.Smile(
        feature_set,
        feature_level,
        num_workers=num_workers,
    )
    fex = audobject.from_yaml_s(
        fex.to_yaml_s(),
        override_args={'num_workers': num_workers},
    )
    assert isinstance(fex, opensmile.Smile)

    # extract from single file

    y_file = fex.process_file(pytest.WAV_FILE)

    # extract from files

    y_files = fex.process_files([pytest.WAV_FILE] * num_files)

    # assertions

    np.testing.assert_equal(np.concatenate([y_file] * num_files),
                            y_files.values)


@pytest.mark.parametrize('feature_set,feature_level', [
    (
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.LowLevelDescriptors,
    ),
    (
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.Functionals,
    ),
])
@pytest.mark.parametrize('index', [
    pd.MultiIndex.from_arrays(
        [
            [pytest.WAV_FILE] * 3,
            pd.to_timedelta([0, 1, 2], unit='s'),
            pd.to_timedelta([1, 2, 3], unit='s'),
        ],
        names=['file', 'start', 'end'],
    ),
])
@pytest.mark.parametrize(
    'num_workers',
    [
        1, 5, None
    ],
)
def test_index(feature_set, feature_level, index, num_workers):

    # create feature extractor

    fex = opensmile.Smile(
        feature_set,
        feature_level,
        num_workers=num_workers,
    )
    fex = audobject.from_yaml_s(
        fex.to_yaml_s(),
        override_args={'num_workers': num_workers},
    )
    assert isinstance(fex, opensmile.Smile)

    # extract from index

    y = fex.process_index(index)

    # extract from files

    files = index.get_level_values(0)
    starts = index.get_level_values(1)
    ends = index.get_level_values(2)
    y_files = fex.process_files(files, starts=starts, ends=ends)

    # assertions

    pd.testing.assert_frame_equal(y, y_files)


@pytest.mark.parametrize('file,feature_set,feature_level', [
    (
        pytest.WAV_FILE,
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.LowLevelDescriptors,
    ),
    (
        pytest.WAV_FILE,
        pytest.CONFIG_FILE,
        opensmile.FeatureLevel.Functionals,
    ),
])
def test_signal(file, feature_set, feature_level):

    # create feature extractor

    fex = opensmile.Smile(feature_set, feature_level)
    fex = audobject.from_yaml_s(fex.to_yaml_s())
    assert isinstance(fex, opensmile.Smile)

    # extract from numpy array

    x, sr = audiofile.read(file, always_2d=True)
    y = fex.process_signal(x, sr)
    y_file = fex.process_file(file)
    y_call = fex(x, sr)
    with pytest.warns(UserWarning):
        y_empty = fex.process_signal(x[0, :10], sr)

    # assertions

    assert y_call.ndim == 3
    assert fex.feature_names == y.columns.to_list()
    np.testing.assert_equal(y.values, y_file.values)
    np.testing.assert_equal(y.values.squeeze(), y_call.squeeze().T)
    assert all(y_empty.isna())



class config:
    r"""Get/set defaults for the :mod:`opensmile` module."""

    CONFIG_ROOT = 'config'
    """Root directory of config files."""

    CONFIG_EXT = '.conf'
    """Extension of config files."""

    EXTERNAL_SOURCE_COMPONENT = 'extsource'
    """Standard component name for external input."""

    EXTERNAL_OUTPUT_COMPONENT = 'extsink'
    """Standard component name for external data output."""

    FILE_INPUT_CONFIG = 'shared/standard_wave_input.conf.inc'
    """Standard config name for wave input from file."""

    EXTERNAL_INPUT_CONFIG = \
        'shared/standard_external_wave_input.conf.inc'
    """Standard config name for external wave input."""

    EXTERNAL_OUTPUT_SINGLE_CONFIG = \
        'shared/standard_external_data_output_single.conf.inc'
    """Standard config name for external data output from a single level."""

    EXTERNAL_OUTPUT_MULTI_CONFIG = \
        'shared/standard_external_data_output_multi.conf.inc'
    """Standard config name for external data output from multiple levels."""

    FILE_OUTPUT_CONFIG = \
        'shared/standard_data_output.conf.inc'
    """Standard config name for external data output."""

    FILE_OUTPUT_CONFIG_NO_LLD_DE = \
        'shared/standard_data_output_no_lld_de.conf.inc'
    """Standard config name for external data output without lld_de level."""



from setuptools import setup


package_data = {
    'opensmile': [
        'core/bin/linux/*',
        'core/bin/osx/*',
        'core/bin/win/*',
        'core/config/compare/*',
        'core/config/egemaps/v01a/*',
        'core/config/egemaps/v01b/*',
        'core/config/egemaps/v02/*',
        'core/config/emobase/*',
        'core/config/gemaps/v01a/*',
        'core/config/gemaps/v01b/*',
        'core/config/shared/*',
    ]
}

setup(
    use_scm_version=True,
    package_data=package_data
)
